﻿using PhysicsUnitsLib;

namespace JumpingPlatformGame {

	struct WorldPoint {
		public Meter X;
		public Meter Y;

		public WorldPoint WithX(Meter newX) => new WorldPoint { X = newX, Y = Y };
		public WorldPoint WithY(Meter newY) => new WorldPoint { X = X, Y = newY };
	}

	abstract class Entity {
		public WorldPoint Location { get; set; }
		public abstract void Update(Second timeDelta);

		public virtual Color Color => Color.Black;
	}

	struct LimitedPositionUpdater {
		public MeterPerSecond Speed { get; set; }

		public Meter LowerBound { get; set; }

		public Meter UpperBound { get; set; }

		public Meter UpdatePosition(Meter position, Second timeDelta, bool reverseOnLower = true, bool reverseOnUpper = true) {
			var distance = Speed * timeDelta;
			var newPosition = position + distance;
			if (newPosition < LowerBound) {
				newPosition = LowerBound;
				if (reverseOnLower) {
					Speed = -1 * Speed;
				}
			}
			if (newPosition > UpperBound) {
				newPosition = UpperBound;
				if (reverseOnUpper) {
					Speed = -1 * Speed;
				}
			}
			return newPosition;
		}
	}

	abstract class MovableEntity : Entity {
		public LimitedPositionUpdater Horizontal;

		public override void Update(Second timeDelta) {
			Location = Location.WithX(Horizontal.UpdatePosition(Location.X, timeDelta));
		}
	}

	abstract class MovableJumpingEntity : MovableEntity {
		public LimitedPositionUpdater Vertical;

		public override void Update(Second timeDelta) {
			base.Update(timeDelta);
			Location = Location.WithY(Vertical.UpdatePosition(Location.Y, timeDelta, reverseOnLower: false));
		}
	}

	class Joe : MovableEntity {
		public override string ToString() => "Joe";
		public override Color Color => Color.Blue;
	}

	class Jack : MovableEntity {
		public override string ToString() => "Jack";
		public override Color Color => Color.LightBlue;
	}

	class Jane : MovableJumpingEntity {
		public override string ToString() => "Jane";
		public override Color Color => Color.Red;
	}

	class Jill : MovableJumpingEntity {
		public override string ToString() => "Jill";
		public override Color Color => Color.Pink;
	}
}